import os

from collections import OrderedDict
from joblib import Parallel, delayed

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
import seaborn as sns

from sklearn.linear_model import LogisticRegression
from sklearn.ensemble import RandomForestClassifier
from sklearn.svm import SVC, LinearSVC
from sklearn.model_selection import StratifiedShuffleSplit
from sklearn.multiclass import OneVsRestClassifier
from sklearn.decomposition import PCA
from sklearn.metrics import roc_auc_score, confusion_matrix, f1_score,\
    precision_recall_curve, PrecisionRecallDisplay, roc_curve, auc
from sklearn.preprocessing import label_binarize

from scipy.stats import ttest_rel

from statsmodels.stats.multitest import multipletests

# ----------------------------------------------------------

class Classifier:

    def __init__(self):

        # Data
        self.data_keys = ['position', 'speed', 'scalars', 'moseq']

        # Selection
        self.y_label = 'y_drug'
        self.fingerprint_mapping = 'dose_mapping'
        self.data_key_choice = 'moseq'

        # Train and plot classifiers and heldout matrix
        self.model_type = "logistic_regression"
        self.C = 0.1
        self.penalty = 'l2'
        self.class_weight = 'balanced'
        self.rbf_decision_function_shape = 'ovr'
        self.rf_n_estimators = 100
        self.validation = True
        self.n_cpu = -1
        self.n_fold = 500
        self.calc_auc = True
        self.f1_average = None
        self.test_size = 0.1
        self.whiten = True
        self.ncomponents = 0.99
        self.order = None
        self.normalize = True
        self.cmap = 'binary_r'
        self.figsize_confusion = (5, 6)
        self.is_fingerprint = True
        self.maxiter = 500

        # Table and plot F1 scores
        self.fscore_average = 'macro'
        self.barplot = False
        self.filter_datatype = None
        self.base_colors = ("#B72025", "#26B34B", "#B4AD33", "#7C449A", "#7C449A")
        self.force_ylim = True

        # PR curve
        self.pr_average = 'weighted'
        self.figsize_pr = (4, 3)

        # Save and show figures
        self.fig_format = 'pdf'
        self.fig_dpi = 300
        self.savefig_dir = None
        self.show_plot = True

    # ----------------------------------------------------------

    def classify_stratified(self, features, labels, labels_for_stratification):

        # ----------------------------------------------------------

        def _classify(features, labels, classifier, train_ind, test_ind):
            classifier.fit(features[train_ind], labels[train_ind])
            if hasattr(classifier, "predict_proba"):
                score = classifier.predict_proba(features[test_ind])
            elif hasattr(classifier, "decision_function"):
                score = classifier.decision_function(features[test_ind])
            else:
                raise ValueError(
                    "Classifier %s has neither a probability or decision score prediction method" % str(classifier))
            pred_labels = classifier.predict(features[test_ind])
            return classifier, labels[test_ind], pred_labels, score

        # ----------------------------------------------------------

        ys = np.unique(labels)

        assert len(features) == len(labels) == len(
            labels_for_stratification), "Features, labels and subject IDs don't match"

        if self.model_type == "logistic_regression":
            model = LogisticRegression
            model_hyparams = dict(penalty=self.penalty, C=self.C, class_weight=self.class_weight)
        elif self.model_type == "svc":
            model = SVC
            model_hyparams = dict(
                kernel='rbf', C=self.C, class_weight=self.class_weight, decision_function_shape=self.rbf_decision_function_shape)
        elif self.model_type == "linearsvc":
            model = LinearSVC
            model_hyparams = dict(C=self.C, class_weight=self.class_weight)
        elif self.model_type == 'rf':
            model = RandomForestClassifier
            model_hyparams = dict(n_estimators=self.rf_n_estimators, class_weight=self.class_weight)
        else:
            raise ValueError(
                "Valid model_type values: ('logistic_regression','svc','linearsvc','rf')")

        if self.whiten:
            proc_features = PCA(self.ncomponents, whiten=True).fit_transform(features)
        else:
            proc_features = features.copy()

        out = Parallel(n_jobs=self.n_cpu)(delayed(_classify)(proc_features,
                                                             labels,
                                                             OneVsRestClassifier(model(**model_hyparams)),
                                                             train_ind, test_ind) for train_ind, test_ind in
                                          StratifiedShuffleSplit(n_splits=self.n_fold, test_size=self.test_size, random_state=0).split(
                                              features, labels_for_stratification))

        classifiers, true_labels, pred_labels, scores = zip(*out)
        aucs = []
        if self.calc_auc:
            ct = 0
            for i in ys:
                aucs.append([])
                for tl, sc in zip(true_labels, scores):
                    aucs[-1].append(roc_auc_score(tl == i, sc[:, ct]))
                ct += 1

        aucs = np.array(aucs)
        return classifiers, true_labels, pred_labels, scores, aucs

    # ----------------------------------------------------------

    def train_and_plot_classifiers(self, fingerprints):

        # ----------------------------------------------------------

        def run_classifier_and_confusion_matrix(fingerprints,
                                                stat='usage',
                                                ylabel='',
                                                ax=None,
                                                fig=None,
                                                normalize=True,
                                                cmap='binary_r',
                                                figsize=(5, 4),
                                                colbar_lim=(0, 1),
                                                is_fingerprint=False,
                                                fingerprint_mapping='dose_mapping',
                                                create_cbar=True,
                                                title=None,
                                                fig_format='pdf',
                                                fig_dpi=300,
                                                savefig_dir=None):

            if is_fingerprint:
                if fingerprint_mapping == 'dose_mapping':
                    X = fingerprints[stat]
                    y = fingerprints['y_dose']
                    mapping = fingerprints['dose_mapping']
                    title = stat
                else:  # drug
                    X = fingerprints[stat]
                    y = fingerprints['y_drug']
                    mapping = fingerprints['drug_mapping']
                    title = stat

            classifiers, true_labels, pred_labels, scores, aucs = self.classify_stratified(X, y, y)

            classifier_rs = {
                'classifiers': classifiers,
                'true_labels': true_labels,
                'pred_labels': pred_labels,
                'scores': scores,
                'aucs': aucs
            }

            if stat == 'usage':
                key = 'MoSeq'
            else:
                key = stat

            plt.set_cmap(cmap)

            # stack the results
            stacked_true_y = np.hstack(true_labels)
            stacked_pred_y = np.hstack(pred_labels)

            # compute confusion matrix
            cm = confusion_matrix(stacked_true_y, stacked_pred_y)

            # normalize
            if normalize:
                cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]

            # plot confusion matrix
            if ax is None:
                save_here = True
                fig, ax = plt.subplots(1, 1, figsize=figsize)
            else:
                save_here = False

            im = ax.imshow(cm, interpolation='nearest', vmin=colbar_lim[0], vmax=colbar_lim[1])

            if create_cbar:
                ax.figure.colorbar(im, ax=ax)
                ax.set(xticks=np.arange(cm.shape[1]),
                       yticks=np.arange(cm.shape[0]),
                       xticklabels=list(mapping.keys()), yticklabels=list(mapping.keys()),
                       ylabel='True label',
                       xlabel='Predicted label')
            else:
                ax.set(xticks=np.arange(cm.shape[1]),
                       yticks=np.arange(cm.shape[0]),
                       xticklabels=list(mapping.keys()), yticklabels=list(mapping.keys()),
                       ylabel=ylabel,
                       xlabel='Predicted label')

            if title:
                ax.set_title(title.capitalize())

            # Rotate the tick labels and set their alignment.
            plt.setp(ax.get_xticklabels(), rotation=45, ha="right", rotation_mode="anchor")

            #fig.tight_layout()

            if save_here:
                plt.savefig(os.path.join(savefig_dir, 'Classifier_confusion_matrix' + key + '.' + fig_format),
                            dpi=self.fig_dpi, bbox_inches='tight')

            return classifier_rs, fig, ax, im

            # ----------------------------------------------------------

        fig, axs = plt.subplots(nrows=1, ncols=len(fingerprints["data_keys"]), sharey=True,
                                figsize=(self.figsize_confusion[0]*len(fingerprints["data_keys"]), self.figsize_confusion[1]),
                                gridspec_kw={'wspace': 0.05, 'hspace': 0.15})

        classifier_results = OrderedDict()
        for i, (ax, data_type) in enumerate(zip(axs, fingerprints["data_keys"])):
            cr = OrderedDict()
            print("Working on %s" % data_type)

            if i == 0:
                ylabel = 'True label'
            else:
                ylabel = ''

            classifier_rs, fig, ax, im = run_classifier_and_confusion_matrix(fingerprints,
                                                                             stat=data_type,
                                                                             cmap=self.cmap,
                                                                             figsize=self.figsize_confusion,
                                                                             normalize=self.normalize,
                                                                             ax=ax, fig=fig,
                                                                             ylabel=ylabel,
                                                                             is_fingerprint=self.is_fingerprint,
                                                                             create_cbar=False,
                                                                             fingerprint_mapping=self.fingerprint_mapping,
                                                                             savefig_dir=self.savefig_dir,
                                                                             fig_format=self.fig_format,
                                                                             fig_dpi=self.fig_dpi)

            cr['true_y'] = classifier_rs['true_labels']
            cr['pred_y'] = classifier_rs['pred_labels']
            cr['scores'] = classifier_rs['scores']
            cr['classifiers'] = classifier_rs['classifiers']
            classifier_results[data_type] = cr

        cbaxes = fig.add_axes([0.95, 0.15, 0.01, 0.65])

        plt.colorbar(im, cax=cbaxes)

        # fig.tight_layout()

        if self.savefig_dir is not None:
            plt.savefig(os.path.join(self.savefig_dir, 'Classifier_confusion_matrix.' + self.fig_format),
                        dpi=self.fig_dpi, bbox_inches='tight')

        if self.show_plot:
            plt.show()

        return classifier_results

    # ----------------------------------------------------------

    def table_and_plot_fscores(self, fingerprints, classifier_results):

        # ----------------------------------------------------------

        def fscores_results_to_dataframe(fingerprints, classifier_results,
                                    fingerprint_mapping='dose_mapping'):

            if fingerprint_mapping == 'dose_mapping':
                unique_fingerprint = 'unique_dose'
                y_label = 'y_dose'
            else:
                unique_fingerprint = 'unique_drug'
                y_label = 'y_drug'

            per_class_metrics = []
            unique_labels = np.unique(fingerprints[y_label])
            for data_type in classifier_results.keys():
                cr = classifier_results[data_type]

                if np.shape(cr['scores'])[2] > 2:
                    true_y = np.hstack(cr['true_y'])
                    pred_y = np.hstack(cr['pred_y'])
                else:
                    true_y = cr['true_y']
                    pred_y = cr['pred_y']
                scores = np.vstack(cr['scores'])
                if scores.ndim == 1:
                    scores = scores[:, None]

                # Binarize the output
                if np.shape(cr['scores'])[2] > 2:
                    true_y = label_binarize(true_y, classes=np.unique(fingerprints[y_label]))
                    pred_y = label_binarize(pred_y, classes=np.unique(fingerprints[y_label]))
                n_classes = true_y.shape[1]

                # Split it back up again
                split_points = np.cumsum([len(s) for s in cr['true_y']])[:-1]
                true_y = np.array_split(true_y, split_points)
                pred_y = np.array_split(pred_y, split_points)
                scores = np.array_split(scores, split_points)

                # Compute performance
                for t, p in zip(true_y, pred_y):
                    for i, drug_name in enumerate(unique_labels):
                        per_class_metrics.append(
                            {

                                'data_type': data_type,
                                'drug': drug_name,
                                'f1_score_' + self.fscore_average: f1_score(t[:, i], p[:, i], average=self.fscore_average),
                            })

            perclassdf = pd.DataFrame(per_class_metrics)
            perclassdf = perclassdf[['data_type'] + [c for c in perclassdf.columns if 'data_type' not in c]]
            dmap = fingerprints[fingerprint_mapping].copy()
            drug_mapper = dict((v, k) for k, v in dmap.items())
            perclassdf['label'] = perclassdf['drug'].map(drug_mapper)

            summarydf = perclassdf.groupby(['drug', 'data_type']).agg(['mean', 'std'])
            summarydf.reset_index(inplace=True)
            summarydf = summarydf.pivot(index='data_type', columns='drug')['f1_score_' + self.fscore_average]

            result = []
            for type_a in perclassdf.data_type.unique():
                for type_b in perclassdf.data_type.unique():
                    if type_a == type_b:
                        continue
                    for idrug in fingerprints[unique_fingerprint]:
                        x = perclassdf[
                            (perclassdf['data_type'] == type_a) & (perclassdf['label'] == idrug)]['f1_score_' + self.fscore_average]
                        y = perclassdf[
                            (perclassdf['data_type'] == type_b) & (perclassdf['label'] == idrug)]['f1_score_' + self.fscore_average]
                        test_result = ttest_rel(x, y)
                        result.append({
                            'drug': idrug,
                            'type_a': type_a,
                            'type_b': type_b,
                            'p': test_result.pvalue,
                            'statistic': test_result.statistic,
                        })

            result_ttest_rel = pd.DataFrame(result)
            _, result_ttest_rel['corrected_p'], _, _ = multipletests(result_ttest_rel.p, alpha=0.01, method='holm')

            return summarydf, perclassdf, result_ttest_rel

        # ----------------------------------------------------------

        def table_f1_results(fingerprints, summarydf, fingerprint_mapping='dose_mapping'):

            new_rows_names = []
            new_row_mean = []
            new_row_std = []
            for s, l in summarydf.keys():
                if s in 'mean':
                    new_rows_names.append([k for k, v in fingerprints[fingerprint_mapping].items() if v == l][0])
                    new_row_mean.append(summarydf[(s, l)].tolist())
                if s in 'std':
                    new_row_std.append(summarydf[(s, l)].tolist())

            new_row_val = []
            for ms, ss in zip(new_row_mean, new_row_std):
                new_row_val.append(['%2.2f±%2.2f' % (m, s) for m, s in zip(ms, ss)])

            new_col_val = np.vstack(new_row_val).T
            new_col_names = [idx for idx in summarydf.index]

            res = pd.DataFrame(dict(zip(new_col_names, new_col_val)), index=new_rows_names)
            # res = res.style.set_table_attributes("style='display:inline'").set_caption('F1 (mean±std)')

            return res

        # ----------------------------------------------------------

        summarydf, perclassdf, result_ttest_rel = fscores_results_to_dataframe(fingerprints, classifier_results,
                                                                               fingerprint_mapping=self.fingerprint_mapping)

        if self.barplot:

            if self.fingerprint_mapping == 'dose_mapping':
                y_label = 'y_dose'
            else:
                y_label = 'y_drug'

            fig = plt.figure(figsize=(20, 4), facecolor='w')
            width = 0.18
            colors = sns.color_palette(self.base_colors)

            mu = summarydf['mean'].loc[fingerprints["data_keys"]].values
            sigma = summarydf['std'].loc[fingerprints["data_keys"]].values

            unique_labels_tmp = np.unique(fingerprints[y_label])

            key_list = list(fingerprints[self.fingerprint_mapping].keys())
            val_list = list(fingerprints[self.fingerprint_mapping].values())
            unique_labels = [key_list[val_list.index(i)] for i in unique_labels_tmp]

            for i, data_type in enumerate(fingerprints["data_keys"]):
                x = np.arange(len(unique_labels)) + i * width
                plt.bar(x, mu[i, :], yerr=sigma[i, :], width=width, ecolor=colors[i], color=colors[i], label=data_type)
            plt.xticks(np.arange(len(unique_labels)) + width * 2, unique_labels, rotation=90, ha='right')
            plt.ylim(0.0, 1.05)
            plt.yticks([0.0, 0.25, 0.5, 0.75, 1.0])
            plt.ylabel('F1')
            plt.legend(loc='center left', bbox_to_anchor=[1, 0.5])
            plt.xlim(-width, len(unique_labels) + width)
            sns.despine()

        else:

            colors = sns.color_palette(self.base_colors)

            if self.filter_datatype is not None:
                perclassdf = perclassdf.loc[perclassdf['data_type'].isin(self.filter_datatype)]

            fig = plt.figure(figsize=(8, 5))
            sns.boxplot(x='label', y='f1_score_' + self.fscore_average, hue='data_type', data=perclassdf, whis=1.5,
                        palette=colors, showfliers=False, dodge=True)
            plt.legend(loc='lower left', bbox_to_anchor=(1.0, 0))
            ticks = [t.get_text() for t in plt.xticks()[1]]
            plt.xticks(range(len(ticks)), ticks, rotation=90);
            plt.ylabel('F1')
            if self.force_ylim:
                axes = plt.gca()
                plt.ylim(-0.1, axes.get_ylim()[1])
            sns.despine()

        plt.tight_layout()

        if self.savefig_dir is not None:
            plt.savefig(os.path.join(self.savefig_dir, 'Fscores.' + self.fig_format),
                        dpi=self.fig_dpi, bbox_inches='tight')

        if self.show_plot:
            plt.show()

        df_results = table_f1_results(fingerprints, summarydf, fingerprint_mapping=self.fingerprint_mapping)

        return df_results, summarydf

    # ----------------------------------------------------------

    def plot_pr_curve_by_measure_averaged(self, fingerprints, classifier_results,
                                      fontsize=12,
                                      fontsize_ticks=10):

        if self.fingerprint_mapping == 'dose_mapping':
            y_label = 'y_dose'
        else:  # drug
            y_label = 'y_drug'

        colors = sns.color_palette(self.base_colors)
        precision_curves = []
        recall_curves = []
        unique_labels_tmp = np.unique(fingerprints[y_label])

        mapping_choice = None
        if y_label == 'y_drug':
            mapping_choice = 'drug_mapping'
        elif y_label == 'y_dose':
            mapping_choice = 'dose_mapping'
        key_list = list(fingerprints[mapping_choice].keys())
        val_list = list(fingerprints[mapping_choice].values())
        unique_labels = [key_list[val_list.index(i)] for i in unique_labels_tmp]

        fig, ax = plt.subplots(figsize=self.figsize_pr)

        f1_data = {}
        auc_results = {}
        for i, data_type in enumerate(classifier_results.keys()):

            cr = classifier_results[data_type]
            if np.shape(cr['scores'])[2] > 2:
                true_y = np.hstack(cr['true_y'])
                pred_y = np.hstack(cr['pred_y'])
            else:
                true_y = cr['true_y']
                pred_y = cr['pred_y']
            scores = np.vstack(cr['scores'])
            if scores.ndim == 1:
                scores = scores[:, None]

            # Binarize the output
            if np.shape(cr['scores'])[2] > 2:
                true_y = label_binarize(true_y, classes=np.unique(fingerprints[y_label]))
                pred_y = label_binarize(pred_y, classes=np.unique(fingerprints[y_label]))

            # Precision and recall
            precision, recall, _ = precision_recall_curve(true_y.ravel(), probas_pred=scores.ravel())
            precision_curves.append(precision)
            recall_curves.append(recall)

            auc_results[data_type] = auc(recall, precision)

            print('--------')
            print('AUC of ' + data_type + ': ' + str(round(auc_results[data_type], 2)))

            per_label_f1 = [f1_score(t, p, average=self.f1_average) for t, p in zip(true_y, pred_y)]
            mean_f1 = np.mean(per_label_f1)
            std_f1 = np.std(per_label_f1) / np.sqrt(len(unique_labels))
            f1_data[data_type] = (mean_f1, std_f1)
            print('F1 score of ' + data_type + ': ' + str(round(mean_f1, 2)) + '±' + str(round(std_f1, 2)))
            print('--------')

        for i, data_type in enumerate(classifier_results.keys()):
            display = PrecisionRecallDisplay(
                recall=recall_curves[i],
                precision=precision_curves[i])

            if len(self.base_colors) >= len(classifier_results.keys()):
                display.plot(ax=ax, name=data_type.capitalize(), color=colors[i])
            else:
                display.plot(ax=ax, name=data_type.capitalize())

        ax.set_xlim([0.0, 1.0])
        ax.set_ylim([0.0, 1.05])
        plt.locator_params(axis="x", nbins=1)
        plt.locator_params(axis="y", nbins=1)
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)

        handles, labels = display.ax_.get_legend_handles_labels()
        ax.legend(handles=handles, labels=labels, bbox_to_anchor=(1.04, 1),
                  loc='upper left', ncol=1, fontsize=fontsize, frameon=False)

        plt.xlabel("Recall", fontsize=fontsize)
        plt.ylabel("Precision", fontsize=fontsize)
        plt.xticks(fontsize=fontsize_ticks)
        plt.xticks(fontsize=fontsize_ticks)
        plt.title(self.pr_average + "-averaged over all classes", fontsize=fontsize)

        plt.tight_layout()

        if self.savefig_dir is not None:
            plt.savefig(os.path.join(self.savefig_dir, 'PR_curve_by_measure_averaged'
                                     + '.' + self.fig_format), dpi=self.fig_dpi, bbox_inches='tight')

        if self.show_plot:
            plt.show()

    # ----------------------------------------------------------

    def plot_pr_curve_by_measure(self, fingerprints, classifier_results):

        if self.fingerprint_mapping == 'dose_mapping':
            y_label = 'y_dose'
        else:  # drug
            y_label = 'y_drug'

        unique_labels_tmp = np.unique(fingerprints[y_label])

        key_list = list(fingerprints[self.fingerprint_mapping].keys())
        val_list = list(fingerprints[self.fingerprint_mapping].values())
        unique_labels = [key_list[val_list.index(i)] for i in unique_labels_tmp]
        for i, data_type in enumerate(classifier_results.keys()):

            cr = classifier_results[data_type]
            true_y = np.hstack(cr['true_y'])
            pred_y = np.hstack(cr['pred_y'])
            scores = np.vstack(cr['scores'])
            if scores.ndim == 1:
                scores = scores[:, None]

            # Binarize the output
            true_y = label_binarize(true_y, classes=np.unique(fingerprints[y_label]))
            pred_y = label_binarize(pred_y, classes=np.unique(fingerprints[y_label]))

            # Compute metrics
            for i in range(len(unique_labels)):
                precision_curve, recall_curve, _ = precision_recall_curve(true_y[:, i], probas_pred=scores[:, i])
                plt.plot(recall_curve, precision_curve, lw=2, label='{}'.format(unique_labels[i]))

            plt.xlabel("Recall", fontsize=22)
            plt.ylabel("Precision", fontsize=22)
            plt.legend(bbox_to_anchor=(1.05, 1.0), loc="upper left", fontsize=12)
            plt.title("{0}: Precision vs. Recall Curve".format(data_type.capitalize()), fontsize=22)
            plt.gca().spines[['top', 'right']].set_visible(False)

            plt.tight_layout()

            if self.savefig_dir is not None:
                plt.savefig(os.path.join(self.savefig_dir, 'PR_curve_by_measure_'
                                         + data_type + '.' + self.fig_format), dpi=self.fig_dpi, bbox_inches='tight')

            if self.show_plot:
                plt.show()

    # ----------------------------------------------------------

    def plot_classifier_au_roc(self, fingerprints, summarydf):

        if self.fingerprint_mapping == 'dose_mapping':
            y_label = 'y_dose'
        else:  # drug
            y_label = 'y_drug'

        unique_labels_tmp = np.unique(fingerprints[y_label])

        key_list = list(fingerprints[self.fingerprint_mapping].keys())
        val_list = list(fingerprints[self.fingerprint_mapping].values())
        unique_labels = [key_list[val_list.index(i)] for i in unique_labels_tmp]

        fig = plt.figure(figsize=(20, 4))
        width = 0.18
        colors = sns.color_palette(self.base_colors)
        mu = summarydf['mean'].loc[fingerprints['data_keys']].values
        sigma = summarydf['std'].loc[fingerprints['data_keys']].values
        for i, data_type in enumerate(fingerprints['data_keys']):
            x = np.arange(len(unique_labels)) + i * width
            plt.bar(x, mu[i, :], yerr=sigma[i, :], width=width, ecolor=colors[i], color=colors[i], label=data_type)
        plt.xticks(np.arange(len(unique_labels)) + width * 2, unique_labels, rotation=90, ha='right')
        plt.ylim(0.4, 1.05)
        plt.yticks([0.5, 0.75, 1.0])
        plt.ylabel('auROC')
        plt.legend(loc='center left', bbox_to_anchor=[1, 0.5])
        plt.xlim(-width, len(unique_labels) + width)
        sns.despine()

        plt.tight_layout()

        if self.savefig_dir is not None:
            plt.savefig(os.path.join(self.savefig_dir, 'Classifier_auROC' +
                                     '.' + self.fig_format), dpi=self.fig_dpi, bbox_inches='tight')

        if self.show_plot:
            plt.show()

    # ----------------------------------------------------------

    def plot_roc_curves(self, fingerprints, classifier_results):

        from scipy import interp

        if self.fingerprint_mapping == 'dose_mapping':
            y_label = 'y_dose'
        else:  # drug
            y_label = 'y_drug'

        if self.data_key_choice == None:
            data_types = fingerprints["data_keys"]
        else:
            data_types = [self.data_key_choice]

        fig = plt.figure(figsize=(5, 3))

        colors = sns.color_palette(self.base_colors)
        for itype, data_type in enumerate(data_types):
            cr = classifier_results[data_type]

            true_y = np.hstack(cr['true_y'])
            pred_y = np.hstack(cr['pred_y'])
            scores = np.vstack(cr['scores'])
            if scores.ndim == 1:
                scores = scores[:, None]

            # Binarize the output
            true_y = label_binarize(true_y, classes=np.unique(fingerprints[y_label]))
            pred_y = label_binarize(pred_y, classes=np.unique(fingerprints[y_label]))
            n_classes = true_y.shape[1]

            # Compute ROC curve and ROC area for each class
            fpr = dict()
            tpr = dict()
            roc_auc = dict()
            for i in range(n_classes):
                fpr[i], tpr[i], _ = roc_curve(true_y[:, i], scores[:, i])
                roc_auc[i] = auc(fpr[i], tpr[i])

            # First aggregate all false positive rates
            all_fpr = np.unique(np.concatenate([fpr[i] for i in range(n_classes)]))

            # Then interpolate all ROC curves at this points
            mean_tpr = np.zeros_like(all_fpr)
            for i in range(n_classes):
                mean_tpr += interp(all_fpr, fpr[i], tpr[i])

            # Finally average it and compute AUC
            mean_tpr /= n_classes

            fpr["macro"] = all_fpr
            tpr["macro"] = mean_tpr
            roc_auc["macro"] = auc(fpr["macro"], tpr["macro"])

            # Plot all ROC curves
            linestyle = '-'
            if 'shuffled' in data_type:
                linestyle = '--'
            plt.plot(fpr["macro"], tpr["macro"], linestyle=linestyle, color=colors[itype],
                     label='{data_type} (area = {0:0.2f})'
                           ''.format(roc_auc["macro"], data_type=data_type))

        plt.legend(loc='center left', bbox_to_anchor=[1, 0.5])
        plt.ylabel("True Positive Rate")
        plt.xlabel("False Positive Rate")
        plt.yticks([0, 0.5, 1.0])
        plt.xticks([0, 0.5, 1.0])
        plt.gca().spines[['top', 'right']].set_visible(False)

        plt.tight_layout()

        if self.savefig_dir is not None:
            plt.savefig(os.path.join(self.savefig_dir, 'ROC_curves' +
                                     '.' + self.fig_format), dpi=self.fig_dpi, bbox_inches='tight')

        if self.show_plot:
            plt.show()


