
import os
import copy
import glob
import h5py
import pickle
import joblib
import numpy as np
from collections import OrderedDict

import seaborn as sns
import matplotlib.pyplot as plt

from moseq2_fingerprint.classifier import *


class Fingerprint:

    def __init__(self):

        # Fingerprint - Data
        self.drug = np.asarray([])
        self.dose = np.asarray([])
        self.drug_class = np.asarray([])
        self.y_highorlow = np.asarray([])
        self.drug_and_dose = np.asarray([])

        self.position = np.asarray([])
        self.speed = np.asarray([])
        self.scalars = np.asarray([])
        self.simplepose = np.asarray([])
        self.moseq = np.asarray([])

        self.data_keys = ['position', 'speed', 'scalars', 'moseq']

        self.unique_dose = np.asarray([])
        self.unique_class = np.asarray([])
        self.unique_drug = np.asarray([])

        self.drug_mapping = OrderedDict([])
        self.class_mapping = OrderedDict([])
        self.dose_mapping = dict()

        self.y_drug = np.asarray([])
        self.yt_drug = np.asarray([])
        self.y_class = np.asarray([])
        self.y_dose = np.asarray([])
        self.yt_class = np.asarray([])


        # Fingerprint - Selection
        self.y_label = 'drug_and_dose'
        self.classifier_fingerprint_mapping = 'dose_mapping'
        self.data_key_choice = 'moseq'

        # Fingerprint - Modification
        self.explained_variance = 0.97

        # Fingerprint - Plotting
        self.cmap_overview = 'cubehelix'
        self.show_class = True
        self.ticks = {
            "Length & Height": np.hstack(((np.linspace(20, 100, 45)), np.linspace(0, 60, 45))),
            "Speed": np.linspace(0, 400, 90),
            "Position": np.linspace(0, 1, 90),
            "MoSeq": np.arange(0, 100)
        }

        # Classifier - Train and plot classifiers
        self.classifier_results_label = 'dose'
        self.classifier_results = None

        self.classifier_model_type = "logistic_regression"
        self.classifier_C = 0.1
        self.classifier_penalty = 'l2'
        self.classifier_class_weight = 'balanced'
        self.classifier_rbf_decision_function_shape = 'ovr'
        self.classifier_rf_n_estimators = 100
        self.classifier_validation = True
        self.classifier_n_cpu = -1
        self.classifier_n_fold = 500
        self.classifier_calc_auc = True
        self.classifier_f1_average = None
        self.classifier_test_size = 0.1
        self.classifier_whiten = True
        self.classifier_ncomponents = 0.99
        self.classifier_order = None
        self.classifier_normalize = True
        self.classifier_cmap = 'binary_r'
        self.classifier_figsize_confusion = (5, 6)
        self.classifier_colbar_lim = (0, 1)
        self.maxiter = 500

        # Classifier - Table and plot F1 scores
        self.summarydf_label = 'dose'
        self.summarydf = None
        self.df_results = None

        self.barplot = False
        self.filter_datatype = None
        self.base_colors = ("#B72025", "#26B34B", "#B4AD33", "#7C449A", "#7C449A")
        self.force_ylim = True

        # Classifier - PR curve
        self.average = 'weighted'
        self.figsize_pr = (5, 4)

        # Save and show figures
        self.fig_format = 'pdf'
        self.fig_dpi = 300
        self.savefig_dir = None
        self.show_plot = True

    # ----------------------------------------------------------

    def save(self, save_path):
        with open(save_path, 'wb') as f:
            pickle.dump(self.__dict__, f)

    # ----------------------------------------------------------

    def load(self, load_path):
        with open(load_path, 'rb') as f:
            dict_tmp = pickle.load(f)
        self.__dict__.update(dict_tmp)

    # ----------------------------------------------------------

    def exclude_groups(self, exclude_groups):

        fingerprint_clone = copy.deepcopy(self.__dict__)

        include_keys = [key for key, value in self.dose_mapping.items() if
                        all(ex_gr not in key for ex_gr in exclude_groups)]
        include_group_idx = [value for key, value in self.dose_mapping.items() if
                             all(ex_gr not in key for ex_gr in exclude_groups)]
        include_idx = np.concatenate(
            [np.where(self.y_dose == idx)[0].tolist() for idx in include_group_idx], dtype='int64')

        fingerprint_clone["position"] = np.array([self.position[idx] for idx in include_idx])
        fingerprint_clone["speed"] = np.array([self.speed[idx] for idx in include_idx])
        fingerprint_clone["simplepose"] = np.array([self.simplepose[idx] for idx in include_idx])
        fingerprint_clone["scalars"] = np.array([self.scalars[idx] for idx in include_idx])
        fingerprint_clone["moseq"] = np.array([self.moseq[idx] for idx in include_idx])
        fingerprint_clone["drug"] = np.array([self.drug[idx] for idx in include_idx])
        fingerprint_clone["dose"] = np.array([self.dose[idx] for idx in include_idx])
        fingerprint_clone["drug_class"] = np.array([self.drug_class[idx] for idx in include_idx])
        fingerprint_clone["y_highorlow"] = np.array([self.y_highorlow[idx] for idx in include_idx])
        fingerprint_clone["drug_and_dose"] = np.array([self.drug_and_dose[idx] for idx in include_idx])
        fingerprint_clone["unique_dose"] = np.array([self.unique_dose[idx] for idx in include_group_idx])
        fingerprint_clone["unique_class"] = [b for a, b in zip([""] + list(fingerprint_clone["drug_class"]), list(fingerprint_clone["drug_class"])) if b != a]
        fingerprint_clone["unique_drug"] = [b for a, b in zip([""] + list(fingerprint_clone["drug"]), list(fingerprint_clone["drug"])) if b != a]
        fingerprint_clone["drug_mapping"] = OrderedDict((i, d) for d, i in enumerate(list(fingerprint_clone["unique_drug"])))
        fingerprint_clone["class_mapping"] = OrderedDict((i, d) for d, i in enumerate(list(fingerprint_clone["unique_class"])))

        fingerprint_clone["y_drug"] = np.array([self.y_drug[idx] for idx in include_idx])
        fingerprint_clone["yt_drug"] = np.r_[0, np.argwhere(np.diff(fingerprint_clone["y_drug"])).ravel()] + 10
        fingerprint_clone["y_class"] = np.array([self.y_class[idx] for idx in include_idx])
        fingerprint_clone["y_dose"] = np.array([self.y_dose[idx] for idx in include_idx])
        yt_class = np.r_[0, np.argwhere(np.diff(fingerprint_clone["y_class"])).ravel(), len(fingerprint_clone["y_class"])]
        fingerprint_clone["yt_class"] = ((yt_class[:-1] + yt_class[1:]) / 2.0).astype('int32')

        fingerprint_clone["dose_mapping"] = dict(
            zip(include_keys, [self.dose_mapping[key] for key in include_keys]))
        items = copy.deepcopy(fingerprint_clone["y_dose"])
        replace_values = [(old, new) for new, (key, old) in
                          enumerate(fingerprint_clone["dose_mapping"].items())]
        dose_mapping_new = []
        for rv in replace_values:
            dose_mapping_new.append(rv[1])
            for index, item in enumerate(items):
                if item == rv[0]:
                    items[index] = rv[1]

        # print('Before:\n' + str([(key, len(value)) for key, value in self.__dict__.items()]))
        # print('----------------------------------------------------------------------------')
        # print('After:\n' + str([(key, len(value)) for key, value in fingerprint_clone.items()]))

        fingerprint_clone["y_dose"] = items
        fingerprint_clone["dose_mapping"] = dict(zip(include_keys, dose_mapping_new))

        self.__dict__.update(fingerprint_clone)

        if len(exclude_groups) > 2:
            print("Successfully removed " + ', '.join(exclude_groups[:-1]) + ' and ' + exclude_groups[-1])
        elif len(exclude_groups) == 2:
            print("Successfully removed " + ' and '.join(exclude_groups))
        else:
            print("Successfully removed " + exclude_groups[0])

    # ----------------------------------------------------------

    def set_max_syllable(self, max_syllable=None, plot=False):

        if max_syllable is None:
            max_syllable = np.argwhere(np.cumsum(self.moseq) <= self.explained_variance).ravel()[-1]


        if plot:
            nsylls = self.moseq.shape[1]

            fig, ax = plt.subplots(1, 1, figsize=(10, 5))
            sns.set_style('ticks')
            ax = sns.lineplot(data=np.mean(self.moseq, axis=0))



            plt.vlines(max_syllable, 0, plt.ylim()[1], colors='k')
            plt.xticks(range(0, nsylls, 10))

            if max_syllable is None:
                plt.title('{exp_var}% Explained variance in {max_sylls} syllables'.format(
                    exp_var=self.explained_variance * 100, max_sylls=max_syllable))
            else:
                plt.title(
                    'Set {max_sylls} syllables'.format(max_sylls=max_syllable))
            plt.ylabel('Syllable usage probability')
            plt.xlabel('Syllable ID')

            sns.despine()
            plt.tight_layout()

            if self.savefig_dir is not None:
                plt.savefig(os.path.join(self.savefig_dir, 'Fingerprint_explained_variance' + '.' + self.fig_format), dpi=self.fig_dpi, bbox_inches='tight')

            if self.show_plot:
                plt.show()

        moseq_clone = copy.deepcopy(self.moseq)

        self.moseq = moseq_clone[:, :max_syllable]

        print("Sucessfully set number of syllables in moseq2-fingerprint to " + str(max_syllable))


    # ----------------------------------------------------------

    def plot_overview_fingerprints(self):

        # ----------------------------------------------------------

        def add_subplot_axes(ax, rect, axisbg='w'):

            fig = plt.gcf()
            box = ax.get_position()
            width = box.width
            height = box.height
            inax_position = ax.transAxes.transform(rect[0:2])
            transFigure = fig.transFigure.inverted()
            infig_position = transFigure.transform(inax_position)
            x = infig_position[0]
            y = infig_position[1]
            width *= rect[2]
            height *= rect[3]  # <= Typo was here
            subax = fig.add_axes([x, y, width, height], facecolor=axisbg)
            x_labelsize = subax.get_xticklabels()[0].get_size()
            y_labelsize = subax.get_yticklabels()[0].get_size()
            x_labelsize *= rect[2] ** 0.5
            y_labelsize *= rect[3] ** 0.5
            subax.xaxis.set_tick_params(labelsize=x_labelsize)
            subax.yaxis.set_tick_params(labelsize=y_labelsize)

            return subax

        # ----------------------------------------------------------

        fig = plt.figure(1, figsize=(20, 15))
        header_fontsize = 25
        gs = plt.GridSpec(1, 9, wspace=0.2, width_ratios=[1, 4, 1, 1, 8, 8, 8, 8, 8])
        axis_scale = 0.8

        tick_locs = np.r_[
            0, np.argwhere(np.diff(self.y_class) != 0).ravel(), len(self.y_class)]
        for t in tick_locs:
            plt.hlines(t, 0, 1, color='k', linewidth=1)
        plt.vlines(0, 0, tick_locs[-1], linewidth=1)

        if self.show_class:
            plt.subplot(gs[0])
            plt.ylim(0, len(self.y_class) - 1)
            plt.gca().invert_yaxis()
            plt.xlim(0, 0.5)
            plt.yticks(self.yt_class, np.array(self.drug_class)[self.yt_class],
                       fontsize=header_fontsize * axis_scale)
            plt.xticks([])
            plt.imshow(self.y_class[:, None], interpolation='nearest', aspect='auto', cmap='Set1')
            plt.title('Class', fontsize=header_fontsize * axis_scale, y=1.02)

        plt.subplot(gs[1])
        plt.title('Drug', fontsize=header_fontsize * axis_scale, y=1.02)
        plt.axis('off')

        plt.subplot(gs[2])
        plt.imshow(self.y_drug[:, None], interpolation='none', cmap='Set2')
        plt.axis('tight')
        plt.yticks(self.yt_drug,
                   [d for i, d in enumerate(self.drug) if i in self.yt_drug],
                   fontsize=header_fontsize * axis_scale)
        plt.xticks([])

        list_dose = [np.where(self.drug_and_dose == i)[0].tolist() for i in
                     np.unique(self.drug_and_dose)]
        center_dose = [l[int(len(l) / 4)] for l in list_dose]

        plt.subplot(gs[3])
        plt.imshow(self.y_dose[:, None], interpolation='none', cmap='Set3')
        plt.axis('tight')
        plt.xticks([])
        plt.yticks(center_dose, np.unique(self.drug_and_dose.tolist()),
                   fontsize=header_fontsize * axis_scale * 0.6, rotation=90)
        plt.tick_params(axis='y', which='major', pad=-20)
        plt.title('Dose', fontsize=header_fontsize * axis_scale, y=1.02)

        plt.gca().spines['top'].set_visible(False)
        plt.gca().spines['right'].set_visible(False)
        plt.gca().spines['bottom'].set_visible(False)
        plt.gca().spines['left'].set_visible(False)
        plt.gca().yaxis.set_ticks_position('none')

        for i, (key, data, scalefn, clim, units) in enumerate([
            ("Position", self.position, np.sqrt, (0, 0.3), "Normalized Distance\nFrom Center"),
            ("Speed", self.speed, np.sqrt, (0, 0.3), "Speed\n(mm/s)"),
            ("Length & Height", self.simplepose, np.sqrt, (0, 0.3), "Length & Height\n(mm)"),
            ("MoSeq", self.moseq, np.sqrt, (0, 0.3), "Syllable ID"),
        ]):
            # Normalize
            data = (data.T / data.sum(1)).T

            plt.subplot(gs[i + 4])
            ax1 = plt.imshow(scalefn(data), clim=clim, interpolation='nearest', cmap=self.cmap_overview)
            _xticks = np.floor(self.ticks[key][::25] / 10.0) * 10.0
            if key == 'Position':
                ax1.axes.set_xticks([0, 36, 72])
                ax1.axes.set_xticklabels([0, 0.4, 0.8])
            else:
                ax1.axes.set_xticks(np.linspace(0, data.shape[1], len(_xticks)))
                ax1.axes.set_xticklabels(map(int, _xticks))
            plt.yticks([])
            plt.axis('tight')
            plt.xlabel(units)
            plt.title(key, y=1.02, fontsize=header_fontsize)
            if key == 'Length & Height':
                plt.vlines(45, 0, len(data), linewidth=1, color='w')
                plt.xticks([0, 60, 45, 89])
            elif key == "MoSeq":
                ax1.axes.set_xticks(range(0, data.shape[1], 10))
                ax1.axes.set_xticklabels(range(0, data.shape[1], 10))
        cbaxes = add_subplot_axes(plt.gca(), [0, -0.06, 1.0, 0.02])
        cb = plt.colorbar(ax1, cax=cbaxes, orientation='horizontal')
        cticks = np.linspace(clim[0], clim[1], 4)
        cb.set_ticks(cticks)
        cb.set_label("Percent usage")
        sns.despine(ax=ax1.axes)

        if self.savefig_dir is not None:
            plt.savefig(os.path.join(self.savefig_dir, 'Fingerprint_overview.' + self.fig_format), dpi=self.fig_dpi, bbox_inches='tight')

        if self.show_plot:
            plt.show()

    # ----------------------------------------------------------

    def plot_selected_groups(self, drug_or_dose, selected_groups, group_labels=None):

        if drug_or_dose == 'dose':
            groups = self.drug_and_dose
        elif drug_or_dose == 'drug':
            groups = self.drug
        else:
            raise ValueError("String does not match. Select either 'drug' or 'dose' for drug_or_dose.")

        sns.set(style="ticks",
                rc={'image.cmap': 'cubehelix', 'font.size': 40, 'pdf.fonttype': 42, 'image.interpolation': 'none'},
                font_scale=1.5)

        if group_labels == None:
            group_labels = selected_groups

        colors = ['#333333', 'r', 'g', 'b', 'orange', 'purple', 'pink',
                  'gold']  # autoselect colors (make a default list of like 10)

        group_inds = [np.array(groups) == g for g in selected_groups]
        fig = plt.figure(figsize=(25, 4))
        gs = plt.GridSpec(1, 5, wspace=0.25)
        for i, (key, data, lim, yticks_, xticks_, units) in enumerate([
            ("Position", self.position, (0, 15), (3, 6, 9, 12), (0, 40, 80, 120),
             'Distance From Center (mm)'),
            ("Speed", self.speed, (0, 5), (1, 2, 3, 4), (0, 75, 150, 225, 300), 'Speed (mm/s)'),
            ("Length & Height", self.simplepose, (0, 15), (3, 6, 9, 12), (0, 100, 200, 300),
             'Length (mm)         Height(mm)'),
            ("MoSeq", self.moseq, (0, 5), (1, 2, 3, 4, 5), None, 'Syllable ID'),
        ]):
            plt.subplot(gs[i])
            for mn, g in enumerate(group_inds):
                try:
                    d = data[g].copy()
                    d = (d.T / d.sum(1)).T
                    plt.plot(100 * d.T, '-', color=colors[mn], alpha=0.1)
                    plt.plot(100 * d.mean(0), '-', color=colors[mn], linewidth=2, label=group_labels[mn])
                except:
                    print(key)

            plt.axis('tight')
            plt.yticks(yticks_)
            if i == 0:
                plt.ylabel('Percent Frames', fontsize=22)
            if i == 1:
                plt.legend(loc='lower left', bbox_to_anchor=(1.0, 0))
            if i == 1:
                plt.xlim(0, 70)
            if i == 2:
                x = [0, 33, 60, 89]
                plt.xticks(x, self.ticks[key][x].astype('int'))
            elif i == 0:
                x = np.searchsorted(self.ticks[key], xticks_)
                # print(x)
                plt.xticks(range(0, max(x) + 1, 30))
            elif i == 1:
                x = np.searchsorted(self.ticks[key], xticks_)
                plt.xticks(x, xticks_)
            if i == 3:
                plt.xlim(0, data.shape[1])

            sns.despine()
            plt.title(key, fontsize=22)
            plt.xlabel(units)

        if self.savefig_dir is not None:
            plt.savefig(os.path.join(self.savefig_dir, 'Fingerprint_selected_groups.' + self.fig_format), dpi=self.fig_dpi, bbox_inches='tight')

        if self.show_plot:
            plt.show()

    # ----------------------------------------------------------

    def train_and_plot_classifiers(self):

        classifier = Classifier()

        classifier.is_fingerprint = True
        classifier.model_type = self.classifier_model_type
        classifier.C = self.classifier_C
        classifier.penalty = self.classifier_penalty
        classifier.class_weight = self.classifier_class_weight
        classifier.rbf_decision_function_shape = self.classifier_rbf_decision_function_shape
        classifier.rf_n_estimators = self.classifier_rf_n_estimators
        classifier.validation = self.classifier_validation
        classifier.n_cpu = self.classifier_n_cpu
        classifier.n_fold = self.classifier_n_fold
        classifier.calc_auc = self.classifier_calc_auc
        classifier.test_size = self.classifier_test_size
        classifier.whiten = self.classifier_whiten
        classifier.ncomponents = self.classifier_ncomponents
        classifier.order = self.classifier_order
        classifier.normalize = self.classifier_normalize
        classifier.cmap = self.classifier_cmap
        classifier.figsize_confusion = self.classifier_figsize_confusion
        classifier.colbar_lim = self.classifier_colbar_lim
        classifier.fingerprint_mapping = self.classifier_fingerprint_mapping
        classifier.fig_format = self.fig_format
        classifier.fig_dpi = self.fig_dpi
        classifier.savefig_dir = self.savefig_dir
        classifier.show_plot = self.show_plot

        classifier_results = classifier.train_and_plot_classifiers(self.__dict__)

        self.classifier_results = classifier_results

        if self.classifier_fingerprint_mapping == 'dose':
            self.classifier_results_label = 'dose'
        else:
            self.classifier_results_label = 'drug'

    # ----------------------------------------------------------

    def table_and_plot_fscores(self):

        classifier = Classifier()

        if self.classifier_results is None or \
            (self.classifier_results_label != self.classifier_results_label):

            classifier.savefig_dir = None
            classifier.show_plot = False

            self.train_and_plot_classifiers()

            if self.classifier_fingerprint_mapping == 'dose':
                self.classifier_results_label = 'dose'
            else:
                self.classifier_results_label = 'drug'

        classifier.fingerprint_mapping = self.classifier_fingerprint_mapping
        classifier.fig_format = self.fig_format
        classifier.fig_dpi = self.fig_dpi
        classifier.savefig_dir = self.savefig_dir
        classifier.show_plot = self.show_plot

        df_results, summarydf = classifier.table_and_plot_fscores(self.__dict__, self.classifier_results)

        self.summarydf = summarydf
        self.df_results = df_results

    # ----------------------------------------------------------

    def plot_pr_curve_by_measure_averaged(self):

        classifier = Classifier()

        if self.classifier_results is None or \
            (self.classifier_results_label != self.classifier_results_label):

            classifier.savefig_dir = None
            classifier.show_plot = False

            self.train_and_plot_classifiers()

            if self.classifier_fingerprint_mapping == 'dose':
                self.classifier_results_label = 'dose'
            else:
                self.classifier_results_label = 'drug'

        classifier.figsize_pr = self.figsize_pr
        classifier.base_colors = self.base_colors
        classifier.average = self.average
        classifier.fingerprint_mapping = self.classifier_fingerprint_mapping
        classifier.fig_format = self.fig_format
        classifier.fig_dpi = self.fig_dpi
        classifier.savefig_dir = self.savefig_dir
        classifier.show_plot = self.show_plot

        classifier.plot_pr_curve_by_measure_averaged(self.__dict__, self.classifier_results)

    # ----------------------------------------------------------

    def plot_pr_curve_by_measure(self):

        classifier = Classifier()

        if self.classifier_results is None or \
            (self.classifier_results_label != self.classifier_results_label):

            classifier.savefig_dir = None
            classifier.show_plot = False

            self.train_and_plot_classifiers()

            if self.classifier_fingerprint_mapping == 'dose':
                self.classifier_results_label = 'dose'
            else:
                self.classifier_results_label = 'drug'

        classifier.fingerprint_mapping = self.classifier_fingerprint_mapping
        classifier.fig_format = self.fig_format
        classifier.fig_dpi = self.fig_dpi
        classifier.savefig_dir = self.savefig_dir
        classifier.show_plot = self.show_plot

        classifier.plot_pr_curve_by_measure(self.__dict__, self.classifier_results)

    # ----------------------------------------------------------

    def plot_classifier_au_roc(self):

        classifier = Classifier()

        if self.summarydf is None or \
            (self.summarydf_label != self.classifier_results_label):

            classifier.savefig_dir = None
            classifier.show_plot = False

            self.table_and_plot_fscores()

            if self.classifier_fingerprint_mapping == 'dose':
                self.summarydf_label = 'dose'
            else:
                self.summarydf_label = 'drug'

        classifier.base_colors = self.base_colors
        classifier.fingerprint_mapping = self.classifier_fingerprint_mapping
        classifier.fig_format = self.fig_format
        classifier.fig_dpi = self.fig_dpi
        classifier.savefig_dir = self.savefig_dir
        classifier.show_plot = self.show_plot

        classifier.plot_classifier_au_roc(self.__dict__, self.summarydf)

    # ----------------------------------------------------------

    def plot_roc_curves(self):

        classifier = Classifier()

        if self.classifier_results is None or \
            (self.classifier_results_label != self.classifier_results_label):

            classifier.savefig_dir = None
            classifier.show_plot = False

            self.train_and_plot_classifiers()

            if self.classifier_fingerprint_mapping == 'dose':
                self.classifier_results_label = 'dose'
            else:
                self.classifier_results_label = 'drug'

        classifier.base_colors = self.base_colors
        classifier.data_key_choice = self.data_key_choice
        classifier.fingerprint_mapping = self.classifier_fingerprint_mapping
        classifier.fig_format = self.fig_format
        classifier.fig_dpi = self.fig_dpi
        classifier.savefig_dir = self.savefig_dir
        classifier.show_plot = self.show_plot

        classifier.plot_roc_curves(self.__dict__, self.classifier_results)
